from datetime import datetime

from openpyxl.styles import Alignment, Font
from openpyxl.worksheet.worksheet import Worksheet

from excel_exporter.configuration.cell_format import CellFormat
from excel_exporter.exporter.character_validation import robust_write_cell


def write_update_time(
    ws: Worksheet,
    update_time: datetime,
    format: CellFormat,
    message: str,
):
    cell = ws.cell(row=1, column=2)
    # Write update time
    full_update_time_message = f'{message} {update_time.strftime("%Y-%m-%d")}'
    robust_write_cell(cell, full_update_time_message)
    # Format update time cell
    alignment = Alignment(
        horizontal=format.horizontal_alignment,
        vertical=format.vertical_alignment,
        wrap_text=format.line_break,
    )
    font = Font(size=format.font_size)
    cell.alignment = alignment
    cell.font = font



from typing import List


def are_strings_grouped(list_of_strings: List[str]) -> bool:
    # First find the number of expected changes
    n_unique_strings = len(set(list_of_strings))
    n_expected_changes = n_unique_strings - 1
    # Then find the real number of changes
    groups_changes = [
        1
        for former, latter in zip(list_of_strings[:-1], list_of_strings[1:])
        if former != latter
    ]
    n_real_changes = sum(groups_changes)
    # Finally compare them
    return n_real_changes == n_expected_changes


# Grouped:
# _list_of_strings = ['a', 'a', 'a', 'b', 'b', 'b', 'c']
# _list_of_strings = ['a', 'a', b']
# _list_of_strings = ['a', 'b', b']
# _list_of_strings = ['a', 'b', c']
# _list_of_strings = ['a', 'b']
# _list_of_strings = ['a', 'a']
# _list_of_strings = ['a']
#
# Not Grouped:
# _list_of_strings = ['a', 'a', 'a', 'b', 'b', 'b', 'a']
# _list_of_strings = ['a', 'b', 'a', 'b', 'b', 'b', 'c']
# _list_of_strings = ['a', 'b', 'a', 'c', 'b', 'c', 'd']
# _list_of_strings = ['a', 'b', 'a', 'b', 'a', 'b', 'a']
# _list_of_strings = ['a', 'b', a']



from typing import Dict

from openpyxl.styles import Alignment, Font, PatternFill
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.worksheet import Worksheet

from excel_exporter.configuration.column import Column as ColumnConfiguration
from excel_exporter.exporter.character_validation import robust_write_cell


def write_columns_header(
    ws: Worksheet,
    columns_config: Dict[str, ColumnConfiguration],
    start_col: int = 2,
    row_number: int = 3,
):
    alignment = Alignment(horizontal='left', vertical='center', wrap_text=True)
    font = Font(bold=True, color='FFFFFF')
    for col_number, col_config in enumerate(
        columns_config.values(), start=start_col
    ):
        # Write column name
        cell = ws.cell(row=row_number, column=col_number)
        robust_write_cell(cell, col_config.column_name)
        # Format column background color in the header
        color = col_config.group.background_color.lstrip('#')
        fill = PatternFill(
            start_color=color, end_color=color, fill_type='solid'
        )
        cell.fill = fill
        # Format column font (bold, color) in the header
        cell.font = font
        # Format column alignment in the header
        cell.alignment = alignment
        # Adjust column width
        col_letter = get_column_letter(col_number)
        ws.column_dimensions[col_letter].width = col_config.column_width



from typing import Dict

from excel_exporter.configuration.cell_format import CellFormat
from excel_exporter.configuration.sheet import Sheet


class ExcelConfiguration:
    def __init__(
        self,
        file_name: str,
        update_message: str,
        cell_formats: Dict[str, CellFormat],
        sheets: Dict[str, Sheet],
    ):
        self.file_name = file_name
        self.update_message = update_message
        self.cell_formats = cell_formats
        self.sheets = sheets

    def __repr__(self, n_indent=2):
        indent = ' ' * n_indent
        s = []
        s.append(f'ExcelConfiguration()')
        s.append(f'{indent}file_name: {self.file_name}')
        s.append(f'{indent}{len(self.cell_formats)} cell_formats')
        s.append(f'{indent}{len(self.sheets)} sheets ')
        return '\n'.join(s)



from datetime import datetime
from typing import Dict, List

from openpyxl.worksheet.worksheet import Worksheet

from excel_exporter.configuration.sheet import Sheet as SheetConfiguration
from excel_exporter.exporter.apply_sheet_configs import apply_sheet_configs
from excel_exporter.exporter.write_groups_header import write_groups_header
from excel_exporter.exporter.write_update_time import write_update_time
from excel_exporter.exporter.write_columns_data import write_columns_data
from excel_exporter.exporter.write_columns_header import write_columns_header


def write_sheet(
    ws: Worksheet,
    ws_data: Dict[str, List],
    ws_config: SheetConfiguration,
    update_time: datetime,
    update_message: str,
):
    # Sheet name
    ws.title = ws_config.sheet_name
    write_update_time(
        ws, update_time, ws_config.update_date_format, update_message
    )
    write_groups_header(ws, ws_config.groups, ws_config.groups_limits())
    write_columns_header(ws, ws_config.columns)
    write_columns_data(ws, ws_data, ws_config.columns)
    apply_sheet_configs(ws, ws_config)



from openpyxl import Workbook
from typing import Dict, List

from excel_exporter.configuration.excel_configuration import ExcelConfiguration
from excel_exporter.exporter.write_sheet import write_sheet


def export_excel(data: List[Dict], config: ExcelConfiguration, update_time):
    wb = Workbook()
    for n, (ws_data, ws_config) in enumerate(
        zip(data, config.sheets.values())
    ):
        ws = wb.active if n == 0 else wb.create_sheet()
        write_sheet(ws, ws_data, ws_config, update_time, config.update_message)
    return wb



from typing import Dict, Tuple

from openpyxl.styles import Alignment, Font, PatternFill
from openpyxl.worksheet.worksheet import Worksheet

from excel_exporter.configuration.group import Group as GroupConfiguration
from excel_exporter.exporter.character_validation import robust_write_cell


def write_groups_header(
    ws: Worksheet,
    groups_config: Dict[str, GroupConfiguration],
    groups_limits: Dict[str, Tuple],
    start_col: int = 2,
    row_number: int = 2,
):
    alignment = Alignment(horizontal='left', vertical='center', wrap_text=True)
    font = Font(bold=True, color='FFFFFF')
    for group_name, group_limits in groups_limits.items():
        group = groups_config[group_name]
        color = group.background_color.lstrip('#')
        fill = PatternFill(
            start_color=color, end_color=color, fill_type='solid'
        )
        group_start_col = group_limits[0] + start_col - 1
        group_end_col = group_limits[1] + start_col - 1
        ws.merge_cells(
            start_row=row_number,
            end_row=row_number,
            start_column=group_start_col,
            end_column=group_end_col,
        )
        cell = ws.cell(row=row_number, column=group_start_col)
        robust_write_cell(cell, group.group_name)
        cell.fill = fill
        cell.alignment = alignment
        cell.font = font



class CellFormat:
    def __init__(
        self,
        format_name: str,
        font_size: int,
        horizontal_alignment: str,
        vertical_alignment: str,
        line_break: bool,
    ):
        self.format_name = format_name
        self.font_size = font_size
        self.horizontal_alignment = horizontal_alignment
        self.vertical_alignment = vertical_alignment
        self.line_break = line_break

    def __repr__(self):
        p = []
        p.append(f'format_name={self.format_name}')
        p.append(f'font_size={self.font_size}')
        p.append(f'horizontal_alignment={self.horizontal_alignment}')
        p.append(f'vertical_alignment={self.vertical_alignment}')
        p.append(f'line_break={self.line_break}')
        parameters = ', '.join(p)
        return f'CellFormat({parameters})'



from openpyxl.cell.cell import Cell, ILLEGAL_CHARACTERS_RE
from openpyxl.utils.exceptions import IllegalCharacterError


def robust_write_cell(cell, value):
    try:
        cell.value = value
    except IllegalCharacterError as e:
        print(f'Found illegal characters on attempt to write on {str(cell)}')
        clean_value = ILLEGAL_CHARACTERS_RE.sub(r'', value)
        illegal_chars = ILLEGAL_CHARACTERS_RE.findall(value)
        illegal_chars = ' '.join(
            '0x' + f'{ord(c):02x}'.upper() for c in set(illegal_chars)
        )
        cell.value = clean_value
        print(f'  - Illegal characters: {illegal_chars}')
        print(f'  - The value was cleaned and written to cell: {clean_value}')
